import torch
from torch import nn

__all__ = ['Bilinear', 'ConvTranspose']


class Bilinear(nn.Module):
    def __init__(self, in_channel: int,
                 out_channel: int):
        super(Bilinear, self).__init__()
        self.Up = nn.Sequential(
            nn.Conv2d(in_channel, out_channel, 1, 1),
            nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
        )

    def forward(self, x):
        return self.Up(x)


class ConvTranspose(nn.Module):
    def __init__(self, in_channel: int,
                 out_channel: int):
        super(ConvTranspose, self).__init__()
        self.Up = nn.ConvTranspose2d(in_channel , out_channel, kernel_size=3, stride=2, padding=1, dilation=1, output_padding=1)

    def forward(self, x):
        return self.Up(x)


import torch
from torch import nn
from tqdm import tqdm
from ..metrics import iou, dice
from ..modules.blocks import *
from ..modules.up import *
from ..modules.down import *
from typing import List, Type
from ..model import binary


__all__ = ['UNet', 'UNet_plus', 'UNet_2plus']

     
class UNet(nn.Module):
    def __init__(self, in_channels: int=1, 
                 out_channels: int=1, 
                 mid_channels: List[int]=[32, 64, 128, 256], 
                 encoder: Type[nn.Module]=VGGBlock,
                 bottom: Type[nn.Module]=None,
                 decoder: Type[nn.Module]=VGGBlock,
                 up: Type[nn.Module]=Bilinear,
                 down: Type[nn.Module]=MaxPooling):
        super().__init__()

        self.enconv = nn.ModuleList()
        self.deconv = nn.ModuleList()
        self.upsamp = nn.ModuleList()
        self.downsamp = nn.ModuleList()

        if bottom:
            self.bridge = bottom(mid_channels[-2], mid_channels[-1])
        else:
            self.bridge = encoder(mid_channels[-2], mid_channels[-1])
        self.final = nn.Conv2d(mid_channels[0], out_channels, kernel_size=1)
        self.down_channels = [in_channels] + mid_channels
        for i in range(len(self.down_channels)-2):
            self.enconv.append(encoder(self.down_channels[i], self.down_channels[i+1]))
        for i in range(len(self.down_channels)-1, 1, -1):
            self.deconv.append(decoder(self.down_channels[i] // 2 + self.down_channels[i-1], self.down_channels[i-1]))
        for i in range(len(self.down_channels)-1, 1, -1):
            self.upsamp.append(up(self.down_channels[i],self.down_channels[i]//2))
        for i in range(len(self.down_channels)-2):
            self.downsamp.append(down(self.down_channels[i]))

    def forward(self, x):
        y_down = []
        y_up = []
        #encoder
        for i in range(len(self.enconv)):
            x = self.enconv[i](x)
            y_down.append(x)
            x = self.downsamp[i](x)
        #bridge
        x = self.bridge(x)
        y_up.append(x)#经过bridge的输出放在y_up的第一位
        #decoder
        for i in range(len(self.deconv)):
            temp = self.deconv[i](torch.cat((y_down[::-1][i], self.upsamp[i](y_up[i])), 1))
            y_up.append(temp)
        return self.final(y_up[-1])
    
    def train_net(self, train_iter, loss=nn.BCEWithLogitsLoss(), optimizer=torch.optim.SGD, num_epochs=40, lr=0.0001, **kwargs):
        updater = optimizer(self.parameters(), lr=lr)
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.to(device)
        self.train()
        epochs = range(num_epochs)
        if 'progress' in kwargs:
            epochs = kwargs['progress'].tqdm(epochs)
        for epoch in epochs:
            loop = tqdm(train_iter)
            for X, y in loop:
                X = X.to(device)
                y = y.to(device)
                y_hat = self(X)
                l = loss(y_hat, y)
                updater.zero_grad()
                l.backward()
                updater.step()
                y_hat = binary(y_hat)
                y = binary(y)
                loop.set_description(f'Epoch [{epoch+1}/{num_epochs}]')
                loop.set_postfix(loss=l.mean().item(), dice=dice(y, y_hat).item(), iou=iou(y, y_hat).item())

class UNet_plus(nn.Module):
    def __init__(self, in_channels: int=1, 
                 out_channels: int=1, 
                 mid_channels: List[int]=[32, 64, 128, 256], 
                 encoder: Type[nn.Module]=VGGBlock,
                 bottom: Type[nn.Module]=None,
                 decoder: Type[nn.Module]=VGGBlock,
                 up: Type[nn.Module]=Bilinear,
                 down: Type[nn.Module]=MaxPooling,
                 deep_supervision: bool=False):
        super().__init__()

        self.enconv = nn.ModuleList()
        self.deconv = nn.ModuleList()
        self.upsamp = nn.ModuleList()
        self.downsamp = nn.ModuleList()
        self.deep_supervision = deep_supervision
        self.out_channels = out_channels
        if bottom:
            self.bridge = bottom(mid_channels[-2], mid_channels[-1])
        else:
            self.bridge = encoder(mid_channels[-2], mid_channels[-1])
        if self.deep_supervision:
            self.final = [nn.Conv2d(mid_channels[0], out_channels, kernel_size=1)] * len(mid_channels)
            self.final = nn.ModuleList(self.final)
        else:
            self.final = nn.Conv2d(mid_channels[0], out_channels, kernel_size=1)
        self.down_channels = [in_channels] + mid_channels

        for i in range(len(self.down_channels)-2):
            self.enconv.append(encoder(self.down_channels[i], self.down_channels[i+1]))
        
        for i in range(2, len(self.down_channels)):
            layer_deconv = nn.ModuleList()
            for j in range(i, 1, -1):
                layer_deconv.append(decoder(self.down_channels[j]//2+self.down_channels[j-1], self.down_channels[j-1]))
            self.deconv.append(layer_deconv)

        for i in range(len(self.down_channels)-2):
            self.downsamp.append(down(self.down_channels[i]))
        
        for i in range(2, len(self.down_channels)):
            layer_up = nn.ModuleList()
            for j in range(i, 1, -1):
                layer_up.append(up(self.down_channels[j],self.down_channels[j]//2))
            self.upsamp.append(layer_up)

    def forward(self, x):
        y = []
        for i in range(len(self.enconv)):
            x = self.enconv[i](x)
            y.append([x])
            x = self.downsamp[i](x)
        y.append([self.bridge(x)])
        
        for i in range(1, len(self.down_channels)-1):
            for j in range(len(self.down_channels)-i-1):
                temp = self.deconv[j+i-1][i-1](torch.cat((y[j][i-1], self.upsamp[j+i-1][i-1](y[j+1][i-1])), 1))
                y[j].append(temp)

        if self.deep_supervision:
            output = []
            for item, f in zip(y[0], self.final):
                output.append(f(item))
            return output
        else:
            return self.final(y[0][-1])
        
    def train_net(self, train_iter, loss=nn.BCEWithLogitsLoss(), optimizer=torch.optim.SGD, num_epochs=40, lr=0.0001, **kwargs):
            updater = optimizer(self.parameters(), lr=lr)
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            self.to(device)
            self.train()
            epochs = range(num_epochs)
            if 'progress' in kwargs:
                epochs = kwargs['progress'].tqdm(epochs)
            for epoch in epochs:
                loop = tqdm(train_iter)
                for X, y in loop:
                    X = X.to(device)
                    y = y.to(device)
                    if self.deep_supervision:
                        y_hats = self(X)
                        l = 0
                        for y_hat in y_hats:
                            l += loss(y_hat, y)
                        l /= len(y_hats)
                    else:
                        y_hat = self(X)
                        l = loss(y_hat, y)
                    updater.zero_grad()
                    l.backward()
                    updater.step()
                    if self.deep_supervision:
                        for i in range(len(y_hats)):
                            y_hats[i] = binary(y_hats[i])
                        y = binary(y)
                        loop.set_postfix(loss=l.mean().item(), dice=[dice(y, y_hat).item() for y_hat in y_hats], iou=[iou(y, y_hat).item() for y_hat in y_hats])
                    else:
                        y_hat = binary(y_hat)
                        y = binary(y)
                        loop.set_description(f'Epoch [{epoch+1}/{num_epochs}]')
                        loop.set_postfix(loss=l.mean().item(), dice=dice(y, y_hat).item(), iou=iou(y, y_hat).item())

    def pruning(self, num_mid_channels):
        self.down_channels = self.down_channels[:num_mid_channels+1]
        self.enconv = self.enconv[:num_mid_channels]
        self.deconv = self.deconv[:num_mid_channels]
        self.upsamp = self.upsamp[:num_mid_channels]
        self.downsamp = self.downsamp[:num_mid_channels]
        if self.deep_supervision:
            self.final = self.final[:num_mid_channels]

    def activate_deep_supervision(self):
        self.deep_supervision = True
        self.final = [nn.Conv2d(self.down_channels[1], self.out_channels, kernel_size=1)] * (len(self.down_channels[1])-1)
        self.final = nn.ModuleList(self.final)
    
    def cancle_deep_supervision(self):
        self.deep_supervision = False
        self.final = self.final[-1]


class UNet_2plus(nn.Module):
    def __init__(self, in_channels: int=1, 
                 out_channels: int=1, 
                 mid_channels: List[int]=[32, 64, 128, 256], 
                 encoder: Type[nn.Module]=VGGBlock,
                 bottom: Type[nn.Module]=None,
                 decoder: Type[nn.Module]=VGGBlock,
                 up: Type[nn.Module]=Bilinear,
                 down: Type[nn.Module]=MaxPooling,
                 deep_supervision: bool=False):
        super().__init__()

        self.enconv = nn.ModuleList()
        self.deconv = nn.ModuleList()
        self.upsamp = nn.ModuleList()
        self.downsamp = nn.ModuleList()
        self.deep_supervision = deep_supervision
        self.out_channels = out_channels
        if bottom:
            self.bridge = bottom(mid_channels[-2], mid_channels[-1])
        else:
            self.bridge = encoder(mid_channels[-2], mid_channels[-1])
        if self.deep_supervision:
            self.final = [nn.Conv2d(mid_channels[0], out_channels, kernel_size=1)] * len(mid_channels)
            self.final = nn.ModuleList(self.final)
        else:
            self.final = nn.Conv2d(mid_channels[0], out_channels, kernel_size=1)
        self.down_channels = [in_channels] + mid_channels

        for i in range(len(self.down_channels)-2):
            self.enconv.append(encoder(self.down_channels[i], self.down_channels[i+1]))
        
        for i in range(2, len(self.down_channels)):
            layer_deconv = nn.ModuleList()
            for j in range(i, 1, -1):
                layer_deconv.append(decoder(self.down_channels[j]//2+self.down_channels[j-1]*(i+1-j), self.down_channels[j-1]))
            self.deconv.append(layer_deconv)

        for i in range(len(self.down_channels)-2):
            self.downsamp.append(down(self.down_channels[i]))
        
        for i in range(2, len(self.down_channels)):
            layer_up = nn.ModuleList()
            for j in range(i, 1, -1):
                layer_up.append(up(self.down_channels[j],self.down_channels[j]//2))
            self.upsamp.append(layer_up)

    def forward(self, x):
        y = []
        for i in range(len(self.enconv)):
            x = self.enconv[i](x)
            y.append([x])
            x = self.downsamp[i](x)
        y.append([self.bridge(x)])
        
        for i in range(1, len(self.down_channels)-1):
            for j in range(len(self.down_channels)-i-1):
                temp = self.deconv[j+i-1][i-1](torch.cat(y[j][0:i]+[self.upsamp[j+i-1][i-1](y[j+1][i-1])], 1))
                y[j].append(temp)

        if self.deep_supervision:
            output = []
            for item, f in zip(y[0], self.final):
                output.append(f(item))
            return output
        else:
            return self.final(y[0][-1])

    def train_net(self, train_iter, loss=nn.BCEWithLogitsLoss(), optimizer=torch.optim.SGD, num_epochs=40, lr=0.0001, **kwargs):
        updater = optimizer(self.parameters(), lr=lr)
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.to(device)
        self.train()
        epochs = range(num_epochs)
        if 'progress' in kwargs:
            epochs = kwargs['progress'].tqdm(epochs)
        for epoch in epochs:
            loop = tqdm(train_iter)
            for X, y in loop:
                X = X.to(device)
                y = y.to(device)
                if self.deep_supervision:
                    y_hats = self(X)
                    l = 0
                    for y_hat in y_hats:
                        l += loss(y_hat, y)
                    l /= len(y_hats)
                else:
                    y_hat = self(X)
                    l = loss(y_hat, y)
                updater.zero_grad()
                l.backward()
                updater.step()
                if self.deep_supervision:
                    for i in range(len(y_hats)):
                        y_hats[i] = binary(y_hats[i])
                    y = binary(y)
                    loop.set_postfix(loss=l.mean().item(), dice=[dice(y, y_hat).item() for y_hat in y_hats], iou=[iou(y, y_hat).item() for y_hat in y_hats])
                else:
                    y_hat = binary(y_hat)
                    y = binary(y)
                    loop.set_description(f'Epoch [{epoch+1}/{num_epochs}]')
                    loop.set_postfix(loss=l.mean().item(), dice=dice(y, y_hat).item(), iou=iou(y, y_hat).item())

    def pruning(self, num_mid_channels):
        self.down_channels = self.down_channels[:num_mid_channels+1]
        self.enconv = self.enconv[:num_mid_channels]
        self.deconv = self.deconv[:num_mid_channels]
        self.upsamp = self.upsamp[:num_mid_channels]
        self.downsamp = self.downsamp[:num_mid_channels]
        if self.deep_supervision:
            self.final = self.final[:num_mid_channels]

    def activate_deep_supervision(self):
        self.deep_supervision = True
        self.final = [nn.Conv2d(self.down_channels[1], self.out_channels, kernel_size=1)] * (len(self.down_channels[1])-1)
        self.final = nn.ModuleList(self.final)
    
    def cancle_deep_supervision(self):
        self.deep_supervision = False
        self.final = self.final[-1]








import torch
from torch import nn
from tqdm import tqdm
from ..metrics import iou, dice
from ..modules.blocks import *
from ..modules.up import *
from ..modules.down import *
from typing import List, Type
from ..model import binary



__all__ = ['FCN32s', 'FCN16s', 'FCN8s']



class FCN32s(nn.Module):
    def __init__(self, in_channels: int=1, 
                 out_channels: int=1, 
                 mid_channels: List[int]=[32, 64, 128, 256], 
                 encoder: Type[nn.Module]=VGGBlock,
                 decoder: Type[nn.Module]=VGGBlock,
                 up: Type[nn.Module]=Bilinear,
                 down: Type[nn.Module]=MaxPooling):
        super().__init__()
        self.enconv = nn.ModuleList()
        self.deconv = nn.ModuleList()
        self.up_bn = nn.ModuleList()
        self.downsamp =  nn.ModuleList()
        self.upsamp = nn.ModuleList()
        self.conv = nn.ModuleList()

        self.down_channels = [in_channels] + mid_channels
        self.final_conv = [nn.Conv2d(self.down_channels[-1], 4096, kernel_size=1),
                    nn.ReLU(inplace=True),
                    nn.Dropout2d(),
                    nn.Conv2d(4096, 4096, kernel_size=1),
                    nn.ReLU(inplace=True),
                    nn.Dropout2d()]
        self.final_conv = nn.Sequential(*self.final_conv)
 
        for i in range(len(self.down_channels)-1):
            self.enconv.append(encoder(self.down_channels[i], self.down_channels[i+1]))
        for i in range(len(self.down_channels)-2):
            self.downsamp.append(down(self.down_channels[i]))
        for i in range(len(self.down_channels)-1, 0, -1):
            self.upsamp.append(up(out_channels , out_channels))
        for i in range(len(self.down_channels)-1, 0, -1):
            self.deconv.append(decoder(out_channels , out_channels))#############
        for i in range(len(self.down_channels)-1, 0, -1):
            self.up_bn.append(nn.BatchNorm2d(out_channels))
        for i in range(len(self.down_channels), 1, -1):
            if i==len(self.down_channels):
                self.conv.append(nn.Conv2d(4096, out_channels, kernel_size=1))
                continue
            self.conv.append(nn.Conv2d(self.down_channels[i-1], out_channels, kernel_size=1))

    def forward(self, x):
        y_down = []
        y_up = []
        #encoder
        for i in range(len(self.enconv)):
            x = self.enconv[i](x)
            x = self.downsamp[i-1](x)
            y_down.append(x)
        #decoder
        x = self.final_conv(x)
        y_down[-1] = x

        for i in range(len(y_down)):
            y_down[i] = self.conv[len(y_down)-i-1](y_down[i])

        for i in range(len(self.down_channels)-1):
            y_up.append(self.up_bn[i](self.deconv[i](self.upsamp[i](y_up[-1]))))
        return y_up[-1] 
        
    def train_net(self, train_iter, loss=nn.BCEWithLogitsLoss(), optimizer=torch.optim.SGD, num_epochs=40, lr=0.0001, **kwargs):#没有修改过
        updater = optimizer(self.parameters(), lr=lr)
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.to(device)
        self.train()
        epochs = range(num_epochs)
        if 'progress' in kwargs:
            epochs = kwargs['progress'].tqdm(epochs)
        for epoch in epochs:
            loop = tqdm(train_iter)
            for X, y in loop:
                X = X.to(device)
                y = y.to(device)
                y_hat = self(X)
                l = loss(y_hat, y)
                updater.zero_grad()
                l.backward()
                updater.step()
                y_hat = binary(y_hat)
                y = binary(y)
                loop.set_description(f'Epoch [{epoch+1}/{num_epochs}]')
                loop.set_postfix(loss=l.mean().item(), dice=dice(y, y_hat).item(), iou=iou(y, y_hat).item())




class FCN16s(nn.Module):
    def __init__(self, in_channels: int=1, 
                 out_channels: int=1, 
                 mid_channels: List[int]=[32, 64, 128, 256], 
                 encoder: Type[nn.Module]=VGGBlock,
                 decoder: Type[nn.Module]=VGGBlock,
                 up: Type[nn.Module]=Bilinear,
                 down: Type[nn.Module]=MaxPooling):
        super().__init__()
        self.enconv = nn.ModuleList()
        self.deconv = nn.ModuleList()
        self.up_bn = nn.ModuleList()
        self.downsamp =  nn.ModuleList()
        self.upsamp = nn.ModuleList()
        self.conv = nn.ModuleList()

        self.down_channels = [in_channels] + mid_channels
        self.final_conv = [nn.Conv2d(self.down_channels[-1], 4096, kernel_size=1),
                    nn.ReLU(inplace=True),
                    nn.Dropout2d(),
                    nn.Conv2d(4096, 4096, kernel_size=1),
                    nn.ReLU(inplace=True),
                    nn.Dropout2d()]
        self.final_conv = nn.Sequential(*self.final_conv)

        for i in range(len(self.down_channels)-1):
            self.enconv.append(encoder(self.down_channels[i], self.down_channels[i+1]))
        for i in range(len(self.down_channels)-2):
            self.downsamp.append(down(self.down_channels[i]))
        for i in range(len(self.down_channels)-1, 0, -1):
            self.upsamp.append(up(out_channels , out_channels))
        for i in range(len(self.down_channels)-1, 0, -1):
            self.deconv.append(decoder(out_channels , out_channels))#############
        for i in range(len(self.down_channels)-1, 0, -1):
            self.up_bn.append(nn.BatchNorm2d(out_channels))
        for i in range(len(self.down_channels), 1, -1):
            if i==len(self.down_channels):
                self.conv.append(nn.Conv2d(4096, out_channels, kernel_size=1))
                continue
            self.conv.append(nn.Conv2d(self.down_channels[i-1], out_channels, kernel_size=1))

    def forward(self, x):
        y_down = []
        y_up = []
        #encoder
        for i in range(len(self.enconv)):
            x = self.enconv[i](x)
            x = self.downsamp[i-1](x)
            y_down.append(x)
        #decoder
        x = self.final_conv(x)
        y_down[-1] = x

        for i in range(len(y_down)):
            y_down[i] = self.conv[len(y_down)-i-1](y_down[i])

        y_up.append(self.up_bn[0](self.deconv[0](self.upsamp[0](y_down[-1]))))
        y_up.append(self.up_bn[1](self.deconv[1](self.upsamp[1](y_up[-1]+y_down[-2]))))
        for i in range(2,len(self.down_channels)-1):
            y_up.append(self.up_bn[i](self.deconv[i](self.upsamp[i](y_up[-1]))))
        return y_up[-1] 
        
    def train_net(self, train_iter, loss=nn.BCEWithLogitsLoss(), optimizer=torch.optim.SGD, num_epochs=40, lr=0.0001, **kwargs):#没有修改过
        updater = optimizer(self.parameters(), lr=lr)
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.to(device)
        self.train()
        epochs = range(num_epochs)
        if 'progress' in kwargs:
            epochs = kwargs['progress'].tqdm(epochs)
        for epoch in epochs:
            loop = tqdm(train_iter)
            for X, y in loop:
                X = X.to(device)
                y = y.to(device)
                y_hat = self(X)
                l = loss(y_hat, y)
                updater.zero_grad()
                l.backward()
                updater.step()
                y_hat = binary(y_hat)
                y = binary(y)
                loop.set_description(f'Epoch [{epoch+1}/{num_epochs}]')
                loop.set_postfix(loss=l.mean().item(), dice=dice(y, y_hat).item(), iou=iou(y, y_hat).item())




class FCN8s(nn.Module):
    def __init__(self, in_channels: int=1, 
                 out_channels: int=1, 
                 mid_channels: List[int]=[32, 64, 128, 256], 
                 encoder: Type[nn.Module]=VGGBlock,
                 decoder: Type[nn.Module]=VGGBlock,
                 up: Type[nn.Module]=Bilinear,
                 down: Type[nn.Module]=MaxPooling):
        super().__init__()
        self.enconv = nn.ModuleList()
        self.deconv = nn.ModuleList()
        self.up_bn = nn.ModuleList()
        self.downsamp =  nn.ModuleList()
        self.upsamp = nn.ModuleList()
        self.conv = nn.ModuleList()

        self.down_channels = [in_channels] + mid_channels
        self.final_conv = [nn.Conv2d(self.down_channels[-1], 4096, kernel_size=1),
                    nn.ReLU(inplace=True),
                    nn.Dropout2d(),
                    nn.Conv2d(4096, 4096, kernel_size=1),
                    nn.ReLU(inplace=True),
                    nn.Dropout2d()]
        self.final_conv = nn.Sequential(*self.final_conv)



        


        for i in range(len(self.down_channels)-1):
            self.enconv.append(encoder(self.down_channels[i], self.down_channels[i+1]))
        for i in range(len(self.down_channels)-2):
            self.downsamp.append(down(self.down_channels[i]))
        for i in range(len(self.down_channels)-1, 0, -1):
            self.upsamp.append(up(out_channels , out_channels))
        for i in range(len(self.down_channels)-1, 0, -1):
            self.deconv.append(decoder(out_channels , out_channels))#############
        for i in range(len(self.down_channels)-1, 0, -1):
            self.up_bn.append(nn.BatchNorm2d(out_channels))
        for i in range(len(self.down_channels), 1, -1):
            if i==len(self.down_channels):
                self.conv.append(nn.Conv2d(4096, out_channels, kernel_size=1))
                continue
            self.conv.append(nn.Conv2d(self.down_channels[i-1], out_channels, kernel_size=1))

    def forward(self, x):
        y_down = []
        y_up = []
        #encoder
        for i in range(len(self.enconv)):
            x = self.enconv[i](x)
            x = self.downsamp[i-1](x)
            y_down.append(x)
        #decoder
        x = self.final_conv(x)
        y_down[-1] = x

        for i in range(len(y_down)):
            y_down[i] = self.conv[len(y_down)-i-1](y_down[i])

        y_up.append(self.up_bn[0](self.deconv[0](self.upsamp[0](y_down[-1]))))
        y_up.append(self.up_bn[1](self.deconv[1](self.upsamp[1](y_up[-1]+y_down[-2]))))
        y_up.append(self.up_bn[2](self.deconv[2](self.upsamp[2](y_up[-1]+y_down[-3]))))
        for i in range(3,len(self.down_channels)-1):
            y_up.append(self.up_bn[i](self.deconv[i](self.upsamp[i](y_up[-1]))))
        return y_up[-1] 
        
    def train_net(self, train_iter, loss=nn.BCEWithLogitsLoss(), optimizer=torch.optim.SGD, num_epochs=40, lr=0.0001, **kwargs):#没有修改过
        updater = optimizer(self.parameters(), lr=lr)
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.to(device)
        self.train()
        epochs = range(num_epochs)
        if 'progress' in kwargs:
            epochs = kwargs['progress'].tqdm(epochs)
        for epoch in epochs:
            loop = tqdm(train_iter)
            for X, y in loop:
                X = X.to(device)
                y = y.to(device)
                y_hat = self(X)
                l = loss(y_hat, y)
                updater.zero_grad()
                l.backward()
                updater.step()
                y_hat = binary(y_hat)
                y = binary(y)
                loop.set_description(f'Epoch [{epoch+1}/{num_epochs}]')
                loop.set_postfix(loss=l.mean().item(), dice=dice(y, y_hat).item(), iou=iou(y, y_hat).item())








import torch
from torch import nn
from torch.nn import functional as F

__all__ = ['VGGBlock', 'ResidualBlock', 'BottleNeck', 'Attention', 'DilatedConv']


class VGGBlock(nn.Module):
    def __init__(self, in_channels: int,
                 out_channels: int,
                 num: int = 2):
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num = num
        super(VGGBlock, self).__init__()
        self.layers_groups = nn.ModuleList()
        channels = [self.out_channels for _ in range(self.num * 2)]
        channels[0] = self.in_channels
        for n in range(self.num):
            layers = nn.Sequential(
                nn.Conv2d(channels[n * 2], channels[n * 2 + 1], kernel_size=3, padding=1),
                nn.BatchNorm2d(channels[n * 2 + 1]),
                nn.ReLU(inplace=True)
            )
            self.layers_groups.append(layers)

    def forward(self, X):
        for layers in self.layers_groups:
            X = layers(X)
        return X

class ResidualBlock(nn.Module):
    def __init__(self, in_channels: int,
                 out_channels: int,
                 num: int = 2):
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num = num
        super(ResidualBlock, self).__init__()
        self.layers_groups = nn.ModuleList()
        self.conv = nn.Conv2d(self.in_channels, self.out_channels, kernel_size=1, padding=0)
        self.RELU = nn.ReLU(inplace=True)
        channels = [self.out_channels for _ in range(self.num * 2)]
        channels[0] = self.in_channels
        for n in range(self.num):
            layers = nn.Sequential(
                nn.Conv2d(channels[n * 2], channels[n * 2 + 1], kernel_size=3, padding=1),
                nn.BatchNorm2d(channels[n * 2 + 1]),
                nn.ReLU(inplace=True),
                nn.Conv2d(channels[n * 2 + 1], channels[n * 2 + 1], kernel_size=3, padding=1),
                nn.BatchNorm2d(channels[n * 2 + 1])
            )
            self.layers_groups.append(layers)

    def forward(self, X):
        for n in range(len(self.layers_groups)):
            if n == 0:
                X = self.RELU(self.layers_groups[n](X) + self.conv(X))
            else:
                X = self.RELU(self.layers_groups[n](X) + X)
        return X

class BottleNeck(nn.Module):
    def __init__(self, in_channels: int,
                 out_channels: int,
                 num: int = 2):
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num = num
        super(BottleNeck, self).__init__()
        self.layers_groups = nn.ModuleList()
        self.conv = nn.Conv2d(self.in_channels, self.out_channels, kernel_size=1, padding=0)
        self.RELU = nn.ReLU(inplace=True)
        channels = [self.out_channels for _ in range(self.num * 2)]
        channels[0] = self.in_channels
        for n in range(self.num):
            layers = nn.Sequential(
                nn.Conv2d(channels[n * 2], channels[n * 2 + 1], kernel_size=1, padding=0),
                nn.BatchNorm2d(channels[n * 2 + 1]),
                nn.ReLU(inplace=True),
                nn.Conv2d(channels[n * 2 + 1], channels[n * 2 + 1], kernel_size=3, padding=1),
                nn.BatchNorm2d(channels[n * 2 + 1]),
                nn.ReLU(inplace=True),
                nn.Conv2d(channels[n * 2 + 1], channels[n * 2 + 1], kernel_size=1, padding=0)
            )
            self.layers_groups.append(layers)

    def forward(self, X):
        for n in range(len(self.layers_groups)):
            if n == 0:
                X = self.RELU(self.layers_groups[n](X) + self.conv(X))
            else:
                X = self.RELU(self.layers_groups[n](X) + X)
        return X

class Attention(nn.Module):
    def __init__(self, in_channels: int,
                 out_channels: int):
        self.in_channels = in_channels
        self.out_channels = out_channels
        super(Attention, self).__init__()
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
    
    def forward(self, x):
        y1 = torch.mean(x, dim=1, keepdim=True)
        y2, _ = torch.max(x, 1, keepdim=True)
        x3 = x * y1
        x4 = x * y2
        x = torch.cat((x3, x4), dim=1)
        x = torch.sigmoid(x)
        return x

class DilatedConv(nn.Module):
    def __init__(self, in_channels: int,
                 out_channels: int):
        self.in_channels = in_channels
        self.out_channels = out_channels
        super(DilatedConv, self).__init__()
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.c1 = nn.Sequential(
            nn.Conv2d(self.in_channels, self.out_channels, kernel_size=3, padding=1, dilation=1),
            nn.BatchNorm2d(self.out_channels),
            nn.ReLU()
        )

        self.c2 = nn.Sequential(
            nn.Conv2d(self.in_channels, self.out_channels, kernel_size=3, padding=3, dilation=3),
            nn.BatchNorm2d(self.out_channels),
            nn.ReLU()
        )

        self.c3 = nn.Sequential(
            nn.Conv2d(self.in_channels, self.out_channels, kernel_size=3, padding=6, dilation=6),
            nn.BatchNorm2d(self.out_channels),
            nn.ReLU()
        )

        self.c4 = nn.Sequential(
            nn.Conv2d(self.in_channels, self.out_channels, kernel_size=3, padding=9, dilation=9),
            nn.BatchNorm2d(self.out_channels),
            nn.ReLU()
        )

        self.c5 = nn.Sequential(
            nn.Conv2d(self.out_channels * 4, self.out_channels, kernel_size=1, padding=0),
            nn.BatchNorm2d(self.out_channels),
            nn.ReLU()
        )

    def forward(self, x):
        x1 = self.c1(x)
        x2 = self.c2(x)
        x3 = self.c3(x)
        x4 = self.c4(x)
        x = torch.cat([x1, x2, x3, x4], axis=1)
        x = self.c5(x)
        return x


import torch
from torch import nn

__all__ = ['MaxPooling', 'AvgPooling', 'ConvPooling']


class MaxPooling(nn.Module):
    def __init__(self, channel: int):
        super(MaxPooling, self).__init__()
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)

    def forward(self, x):
        return self.pool(x)

class AvgPooling(nn.Module):
    def __post_init__(self, channel: int):
        super(AvgPooling, self).__init__()
        self.pool = nn.AvgPool2d(kernel_size=2, stride=2)

    def forward(self, x):
        return self.pool(x)

class ConvPooling(nn.Module):
    def __post_init__(self, channel: int):
        super(ConvPooling, self).__init__()
        self.conv = nn.Conv2d(channel, channel, kernel_size=2, stride=2, padding=0)

    def forward(self, x):
        return self.conv(x)


from typing import Union
import numpy as np
import torch

def dice(a: Union[np.ndarray, torch.Tensor], b: Union[np.ndarray, torch.Tensor]) -> float:
    return 2 * (a.bool() & b.bool()).long().sum() / (a.sum() + b.sum())

def iou(a: Union[np.ndarray, torch.Tensor], b: Union[np.ndarray, torch.Tensor]) -> float:
    return (a.bool() & b.bool()).long().sum() / (a.bool() | b.bool()).long().sum()


import os
import torch
from torch import nn
from torch.utils import data
import torchvision.transforms as transforms
from zipfile import ZipFile
import cv2


class ImageDataset(data.Dataset):
    def __init__(self, path: str, color: str = 'gray', mode: str = 'train'):
        self.path = path
        self.color = color
        self.mode = mode
        self.name = os.listdir(self.path + '/image')
        self.transform = transforms.ToTensor()
    
    def __getitem__(self, idx):
        if self.mode == 'train':
            if self.color == 'gray':
                image = cv2.imread(self.path + '/image/' + self.name[idx], cv2.IMREAD_GRAYSCALE)
                mask = cv2.imread(self.path + '/mask/' + self.name[idx], cv2.IMREAD_GRAYSCALE)
            elif self == 'RGB':
                image = cv2.imread(self.path + '/image/' + self.name[idx])
                mask = cv2.imread(self.path + '/mask/' + self.name[idx])
            image = self.transform(image)
            mask = self.transform(mask)
            return image, mask
        elif self.mode == 'test':
            if self.color == 'gray':
                image = cv2.imread(self.path + '/image/' + self.name[idx], cv2.IMREAD_GRAYSCALE)
            elif self == 'RGB':
                image = cv2.imread(self.path + '/image/' + self.name[idx])
            image = self.transform(image)
            return image

    def __len__(self):
        return len(self.name)


def show_img(tensor):
    unloader = transforms.ToPILImage()
    image = tensor.clone()
    image = image.squeeze(0)
    image = unloader(image)
    return image

def show_mask(img, mask, transparency=0.5):
    color = torch.tensor([30/255, 144/255, 255/255]).reshape(-1, 1, 1)
    transform = transforms.ToTensor()
    img = transform(img)
    mask = transform(mask)
    front = img * mask
    back = img - front
    result = back + (mask * color * (1 - transparency) + front * transparency)
    return show_img(result)




from typing import Type
import torch
import torch.nn as nn

def binary(tensor: Type[torch.Tensor], threshold: float=0.5) -> Type[torch.Tensor]:
    func = nn.Sigmoid()
    tensor = func(tensor)
    tensor[tensor >= threshold] = 1
    tensor[tensor < threshold] = 0
    return tensor


from distutils.core import setup
from setuptools import find_packages

with open("README.md", "r") as f:
  long_description = f.read()

setup(name='easymed', 
      version='0.0.6',
      description='One stop medical toolkit',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='EasyMed in BUPT',
      author_email='bool1020@bupt.edu.cn',
      url='https://github.com/EasyMed2023',
      install_requires=[
        'tqdm',
        'torch',
        'opencv-python'
        ],
      license='MIT License',
      packages=find_packages(),
      platforms=["all"],
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.10',
      ],
      python_requires='>=3.7',
      )